use core::cmp::max;

use avr_device::{
    atmega32u4::{PLL, USB_DEVICE},
    interrupt::{free, CriticalSection, Mutex},
};
use usb_device::{
    bus::{PollResult, UsbBus},
    endpoint::{EndpointAddress, EndpointType},
    Result as UsbResult, UsbDirection, UsbError,
};

#[allow(unused)]
#[derive(Default, Copy, Clone)]
pub(crate) struct USBEndpoint {
    is_allocated: bool,
    size: u8,
    ep_type: u8,
    ep_dir: bool,
    banks: u8,
}

const ENDPOINTS_ALLOC_LAYOUT: [u16; 7] = [64, 256, 64, 64, 64, 64, 64];

impl USBEndpoint {
    #[inline]
    fn set_type(&mut self, ep_type: EndpointType) {
        self.ep_type = match ep_type {
            EndpointType::Control => 0, // 0 = 0b00
            EndpointType::Isochronous {
                synchronization: _,
                usage: _,
            } => 1, // 1 = 0b01
            EndpointType::Bulk => 2,    // 2 = 0b10
            EndpointType::Interrupt => 3, // 3 = 0b11
        };
    }

    #[inline]
    fn set_dir(&mut self, dir: UsbDirection) {
        self.ep_dir = match dir {
            UsbDirection::In => true,
            UsbDirection::Out => false,
        };
    }

    #[inline]
    fn set_size(&mut self, size: u16) {
        self.size = match size {
            8 => 0b000,
            16 => 0b001,
            32 => 0b010,
            64 => 0b011,
            128 => 0b100,
            256 => 0b101,
            512 => 0b110,
            _ => unreachable!(),
        }
    }
}

pub struct UsbDevice<const L: usize> {
    pll: Mutex<PLL>,
    usb: Mutex<USB_DEVICE>,
    ep_table: [USBEndpoint; L],
    dpram_already_used: u16,
}

const DPRAM_SIZE: u16 = 832;

impl<const L: usize> UsbDevice<L> {
    #[inline]
    pub fn new(pll: PLL, usb: USB_DEVICE) -> Self {
        let (pll, usb) = (Mutex::new(pll), Mutex::new(usb));
        let ep_table: [USBEndpoint; L] = [Default::default(); L];
        Self {
            pll,
            usb,
            ep_table,
            dpram_already_used: 0,
        }
    }

    pub fn select_endpoint(
        &mut self,
        cs: CriticalSection<'_>,
        endpoint_index: usize,
    ) -> Result<(), UsbError> {
        let usb = self.usb.borrow(cs);
        let endpoint_index = endpoint_index as u8;

        if endpoint_index > 6 {
            return Err(UsbError::InvalidEndpoint);
        }

        usb.uenum.write(|w| w.bits(endpoint_index));

        if usb.uenum.read().bits() != endpoint_index {
            return Err(UsbError::InvalidEndpoint);
        }

        Ok(())
    }

    pub fn configure_endpoint(
        &mut self,
        cs: CriticalSection<'_>,
        endpoint_index: usize,
    ) -> Result<(), UsbError> {
        match self.select_endpoint(cs, endpoint_index) {
            Ok(_) => {
                let usb = self.usb.borrow(cs);
                let current_endpoint = self.ep_table[endpoint_index];

                // Clear interrupt. //
                
                usb.udint.modify(|_, w| w.eorsti().clear_bit());

                // Enable endpoint. //

                usb.ueconx.modify(|_, w| w.epen().set_bit());

                // Set markered endpoint parameters to uecfg0x/1x register. //

                usb.uecfg0x.modify(|_, w| {
                    w.epdir()
                        .bit(current_endpoint.ep_dir)
                        .eptype()
                        .bits(current_endpoint.ep_type)
                });

                usb.uecfg1x.modify(|_, w| {
                    w.epbk()
                        .bits(current_endpoint.banks)
                        .epsize()
                        .bits(current_endpoint.size)
                        .alloc()
                        .bit(current_endpoint.is_allocated)
                });

                if !usb.uesta0x.read().cfgok().bit() {
                    Err(UsbError::EndpointOverflow)
                } else {
                    Ok(())
                }
            }
            Err(exception) => Err(exception),
        }
    }
}

